



pytest --pyargs --doctest-modules --doctest-ignore-import-errors datashader
IF %ERRORLEVEL% NEQ 0 exit /B 1
datashader copy-examples --path=. --force
IF %ERRORLEVEL% NEQ 0 exit /B 1
datashader fetch-data --path=. --datasets small.yml
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
